import { NextRouter, SingletonRouter } from 'next/router'
import type { UrlObject } from 'url'

import type { Url } from '../types'
import { fileUrlToUrl } from './fileUrlToUrl'
import { getLocale } from './getLocale'
import { getNtrData } from './ntrData'
import { removeLangPrefix } from './removeLangPrefix'
import { urlToFileUrl } from './urlToFileUrl'

export const translatePushReplaceArgs = ({
  router,
  url,
  as,
  locale,
}: {
  router: NextRouter | SingletonRouter
  url: Url
  as?: Url
  locale?: string | false
}) => {
  if (as) {
    return { url, as, locale }
  }

  let newLocale = getLocale(router, locale)
  const locales = router.locales || getNtrData().locales
  const unprefixedUrl = typeof url === 'string' ? removeLangPrefix(url) : url
  const urlLocale = typeof url === 'string' && unprefixedUrl !== url ? url.split('/')[1] : undefined

  // propLocale === false if opted-out of automatically handling the locale prefixing
  // Cf. https://nextjs.org/docs/advanced-features/i18n-routing#transition-between-locales
  if (locale === false && urlLocale && locales.includes(urlLocale)) {
    newLocale = urlLocale
  }

  let parsedUrl: UrlObject | URL | string | undefined
  let translatedUrl: string | undefined

  /**
   * Href can be:
   * - an external url
   * - a correct file url
   * - a wrong file url (not matching any page)
   * - a correct translated url
   * - a wrong translated url
   */
  try {
    translatedUrl = fileUrlToUrl(unprefixedUrl, newLocale)
    // Url is a correct file url
    parsedUrl = unprefixedUrl
  } catch {
    parsedUrl = urlToFileUrl(unprefixedUrl, urlLocale || newLocale)
    if (parsedUrl) {
      try {
        translatedUrl = fileUrlToUrl(parsedUrl, newLocale)
        // Href is a correct translated url
      } catch {
        // Href is a wrong file url or an external url
      }
    } else {
      // Href is a wrong url or an external url
    }
  }

  return { url: parsedUrl || url, as: translatedUrl, locale: newLocale }
}
